#!/usr/bin/env python3

import os
import sys

from samba.samba3 import param as s3param
from samba.dcerpc import security
from samba.auth_util import system_session_unix
from samba.netcmd.ntacl import get_local_domain_sid
from samba.ntacls import getntacl, setntacl
from samba.netcmd.main import cmd_sambatool
cmd = cmd_sambatool()

import traceback


sddl = sys.argv[1]
file = sys.argv[2].rstrip('/')
chown = (int)(sys.argv[3])
inherit = False
force_cover = (int)(sys.argv[4])
s3conf = s3param.get_context()
# This step is to lock one db
s3conf.load_default()

lp = s3conf
domsid = get_local_domain_sid(lp)
session_info = system_session_unix()


SEC_DESC_DACL_AUTO_INHERITED = 0x0400
SEC_DESC_DACL_PROTECTED = 0x1000
CHOWN_FILE = 0x1
CHOWN_FOLDER = 0x2
# File
ACE_FLAG_OI = 0x1
# Dir
ACE_FLAG_CI = 0x2
# Inherit
ACE_FLAG_ID = 0x10

ACE_TYPE_ALLOW = 0x0
ACE_TYPE_DENY = 0x1


"""
def setntacl(lp, path, new_sd,domsid,session_info, use_ntvfs):
    print("set", path, new_sd.as_sddl())
"""


def ace_has(aces, ace):
    """
    If ace exists in aces.

    Args:
        aces (list[security.ace]): The source aces
        ace (security.ace): The target ace
    """
    for ace_ori in aces:
        if ace_ori.access_mask == ace.access_mask \
            and ace_ori.flags == ace.flags \
            and ace_ori.object == ace.object \
            and (str)(ace_ori.trustee) == (str)(ace.trustee) \
            and ace_ori.type == ace.type:
            return True
    return False


def ace_rules(aces):
    """
    Remove dupe ace in aces and sort them as deny must be before allow.

    Args:
        aces (list[security.ace]): The original aces
    """
    allow_aces = []
    deny_aces = []
    allow_aces_inherit = []
    deny_aces_inherit = []
    for ace in aces:
        if ace.type == ACE_TYPE_ALLOW:
            if ace.flags & ACE_FLAG_ID:
                if not ace_has(allow_aces_inherit, ace):
                    allow_aces_inherit.append(ace)
            else:
                if not ace_has(allow_aces, ace):
                    allow_aces.append(ace)
        elif ace.type == ACE_TYPE_DENY:
            if ace.flags & ACE_FLAG_ID:
                if not ace_has(deny_aces_inherit, ace):
                    deny_aces_inherit.append(ace)
            else:
                if not ace_has(deny_aces, ace):
                    deny_aces.append(ace)

    return deny_aces + allow_aces + deny_aces_inherit + allow_aces_inherit


def sddl_dacl_sort(sd):
    """
    Sort aces in dacl since deny must be before allow.

    Args:
        sd (security.descriptor): The sd.
    """
    new_sd = security.descriptor.from_sddl(
        "O:" + (str)(sd.owner_sid) + "G:" + (str)(sd.group_sid) + "D:",
        domsid)

    new_sd.type = sd.type

    for ace in ace_rules(sd.dacl.aces):
        new_sd.dacl_add(ace)

    return new_sd


def wash_inherit(sd):
    """
    Remove inherit ace.

    Args:
        sd (security.descriptor): The sd.
    """
    new_sd = security.descriptor.from_sddl(
        "O:" + (str)(sd.owner_sid) + "G:" + (str)(sd.group_sid) + "D:",
        domsid)

    new_sd.type = sd.type

    for ace in sd.dacl.aces:
        if not ace.flags & ACE_FLAG_ID:
            new_sd.dacl_add(ace)

    return new_sd


def merge_inherit(sd, parent_sd, chown = False):
    """
    Merge the sd with inherited parent_sd

    Args:
        sd (security.descriptor): The old sd.
        parent_sd (security.descriptor): The parent sd to inherit.
        chown (bool, optional): If change the owner. Defaults to False.

    Return:
        merged sd
    """
    if chown:
        new_sd_str = "O:" + (str)(parent_sd.owner_sid) + "G:" + (str)(parent_sd.group_sid)
    else:
        new_sd_str = "O:" + (str)(sd.owner_sid) + "G:" + (str)(sd.group_sid)

    new_sd_str += "D:"

    new_sd = security.descriptor.from_sddl(new_sd_str, domsid)

    new_sd.type = parent_sd.type

    aces = []
    for ace in sd.dacl.aces:
        if not ace.flags & ACE_FLAG_ID:
            aces.append(ace)
    for ace in parent_sd.dacl.aces:
        aces.append(ace)

    for ace in ace_rules(aces):
        new_sd.dacl_add(ace)

    return sddl_dacl_sort(new_sd)


def update_inherit_sddl(parent_folder_sddl, file_sddl, folder_sddl):
    """
    Update the sd info to fit objects in folder.

    Args:
        parent_folder_sddl (security.descriptor): The sddl of parent folder.
        file_sddl (security.descriptor): The sddl for files to inherit.
        folder_sddl (security.descriptor): The sddl for folders to inherit.
    """
    file_sddl_new = clear_inherit(file_sddl)
    folder_sddl_new = clear_inherit(folder_sddl)

    for ace in parent_folder_sddl.dacl.aces:
        if ace.flags & ACE_FLAG_OI:
            file_sddl_new.dacl_add(ace)
            file_sddl_new.dacl.aces[-1].flags = ACE_FLAG_ID
        if ace.flags & ACE_FLAG_CI:
            folder_sddl_new.dacl_add(ace)
            folder_sddl_new.dacl.aces[-1].flags = folder_sddl_new.dacl.aces[-1].flags | ACE_FLAG_ID

    return file_sddl_new, folder_sddl_new


def setntacl_inherit(lp, path, file_sddl, folder_sddl, session_info, chown = 0):
    """
    Set the child files and folders
    """
    return_code = 0
    try:
        sd = getntacl(lp, path, session_info, service=None, direct_db_access=False)
        if not (inherited(sd.type) or force_cover):
            return return_code

        if os.path.islink(path):
            pass

        elif os.path.isfile(path):
            new_sd = merge_inherit(sd, file_sddl, (bool)(chown & CHOWN_FILE))
            if not sd.as_sddl() == new_sd.as_sddl():
                setntacl(lp, path, new_sd, domsid, session_info, use_ntvfs=False)

        elif os.path.isdir(path):
            new_sd = merge_inherit(sd, folder_sddl, (bool)(chown & CHOWN_FOLDER))
            if not sd.as_sddl() == new_sd.as_sddl():
                setntacl(lp, path, new_sd, domsid, session_info, use_ntvfs=False)

                # Update inherit sddl info
                file_sddl_new, folder_sddl_new = update_inherit_sddl(new_sd, file_sddl, folder_sddl)

                for file in os.listdir(path):
                    return_code = return_code | setntacl_inherit(lp, os.path.join(path, file), file_sddl_new, folder_sddl_new, session_info, chown)

    except Exception:
        print(traceback.format_exc())
        return_code = 0x2

    return return_code

def clear_inherit(sddl):
    """
    Clear inherited sddl.

    Args:
        sddl (security.descriptor): The sddl for files to inherit.
    """
    sddl_str = sddl.as_sddl()
    sddl_old = security.descriptor.from_sddl(sddl_str, domsid)

    sddl_strlen = sddl_str.find('(')
    if sddl_strlen != -1:
        sddl_new = security.descriptor.from_sddl(sddl_str[:sddl_strlen], domsid)
    else:
        sddl_new = security.descriptor.from_sddl(sddl_str, domsid)

    for ace in sddl_old.dacl.aces:
        if not (ace.flags & ACE_FLAG_ID):
            sddl_new.dacl_add(ace)

    return sddl_new

def merge_parent(file, sd):
    """
    Merge parent sddl.

    Args:
        file (str): The path of the file.
        sd (security.descriptor): The sddl for files to inherit.
    """
    new_sd = sd
    if not os.path.ismount(file):
        try:
            parent_path = os.path.dirname(file)
            if not os.path.ismount(parent_path):
                parent_sd = getntacl(lp, parent_path, session_info, service=None, direct_db_access=False)
                file_sddl, folder_sddl = update_inherit_sddl(parent_sd, sd, sd)
                new_sd = file_sddl if os.path.isfile(file) else folder_sddl

        except Exception as e:
            pass
    return new_sd


def inherited(type):
    """
    Return if this is inherited
    """
    return (type & SEC_DESC_DACL_AUTO_INHERITED) and not (type & SEC_DESC_DACL_PROTECTED)


def main():
    global inherit
    retval = 0

    if os.path.islink(file):
        sys.exit(1)

    sd = clear_inherit(security.descriptor.from_sddl(sddl, domsid))
    if inherited(sd.type):
        inherit = True
        sd = merge_parent(file, sd)
    else:
        sd = wash_inherit(sd)

    try:
        sd = sddl_dacl_sort(sd)
        setntacl(lp, file, sd, domsid, session_info, use_ntvfs=False)
    except SystemExit as e:
        print("SystemExit main")
        retval = e.code
    except Exception as e:
        cmd.show_command_error(e)
        retval = 1

    if os.path.isdir(file) and inherited(sd.type):
        try:
            raw_sddl = "O:" + (str)(sd.owner_sid) + "G:" + (str)(sd.group_sid) + "D:AI"
            file_sddl = security.descriptor.from_sddl(raw_sddl, domsid)
            folder_sddl = security.descriptor.from_sddl(raw_sddl, domsid)
            file_sddl, folder_sddl = update_inherit_sddl(sd, file_sddl, folder_sddl)

            for files in os.listdir(file):
                retval = retval | setntacl_inherit(lp, os.path.join(file, files), file_sddl, folder_sddl, session_info, chown)
        except SystemExit as e:
            print("SystemExit inherit")
            retval = e.code
        except Exception as e:
            cmd.show_command_error(e)
            retval = retval | 1

    sys.exit(retval)

if __name__ == '__main__':
    main()
